/* global nsVmap, ol, oVmap, goog, angular, vitisApp, bootbox */

/**
 * @author: Anthony Borghi
 * @Description: Fichier contenant la classe nsVmap.MapCompare
 * cette classe permet l'initialisation de la carte
 */

goog.provide('nsVmap.MapCompare');

goog.require('oVmap');

goog.require('ol.Map');
goog.require('ol.View');
goog.require('MapJSON');
goog.require('ol.Overlay');
goog.require('ol.source.Vector');
goog.require('ol.layer.Vector');
goog.require('goog.async.AnimationDelay');

/**
 * @classdesc
 * Class {@link nsVmap.MapCompare}: initializes the map to compare
 *
 * @constructor
 * @export
 */
nsVmap.MapCompare = function () {
    oVmap.log("nsVmap.MapCompare");

    var this_ = this;

    /**
     * Map layers
     * @type {Array.<ol.layer>}
     * @private
     */
    this.olLayers_ = [];

    /**
     * Map view
     * @type {ol.View}
     * @private
     */
    this.olView_ = new ol.View({
        center: [0, 0],
        zoom: 2
    });

    /**
     * Map
     * @type {ol.map}
     * @private
     */
    this.oOpenLayersMap_ = new ol.Map({
        layers: this.olLayers_,
        view: this.olView_,
        moveTolerance: 5
    });

    this.oOpenLayersMap_.getViewport().setAttribute('id', 'map1Compare');
    /**
     * Contient les évènements ajoutés sur la carte par la méthode addEventOnMap ou setEventOnMap
     * @type {array}
     * @private
     */
    this.vmapEvents_ = [];

    /**
     * Contient les évènements ajoutés sur la carte par la méthode addDrawInteraction ou setDrawInteraction
     * @type {array}
     * @private
     */
    this.vmapInteractions_ = [];

    /**
     * Contient les tooltips ajoutés par addTooltip
     * @type {array}
     * @private
     */
    this.vmapOverlays_ = [];

    /**
     * Initialisé dans le controlleur
     * @type {nsVmap.Map.MapTooltip}
     * @private
     */
    this.vmapTooltip_ = {};
    // Ajoute les couches à reprojeter en cas de changement de carte dans layersToTransform_
    this.layersToTransform_ = [];
}

/**
 * nsVmap.Map.oOpenLayersMap_ getter
 * @return {ol.map} Used map
 * @export
 * @api experimental
 */
nsVmap.MapCompare.prototype.getOLMap = function () {
    return this.oOpenLayersMap_;
};

/**
 * nsVmap.Map.olView_ setter
 * @param {ol.View} view View to use
 * @export
 * @api experimental
 */
nsVmap.MapCompare.prototype.setOpenLayersView = function (view) {
    this.olView_ = view;
    this.oOpenLayersMap_.setView(this.olView_);
};

/**
 * Use this function to add a new vector layer
 * @param {object} opt_options
 * @param {object} opt_options.style
 * @param {object} opt_options.extent
 * @param {object} opt_options.minResolution
 * @param {object} opt_options.maxResolution
 * @param {object} opt_options.opacity
 * @param {object} opt_options.properties
 * @param {object} opt_options.visible
 * @param {object} opt_options.zIndex
 * @returns {undefined}
 */
nsVmap.MapCompare.prototype.addVectorLayer = function (opt_options) {
    oVmap.log('nsVmap.MapCompare.prototype.addVectorLayer');

    var vectorLayer = new ol.layer.Vector({
        map: this.oOpenLayersMap_,
        source: new ol.source.Vector({
            useSpatialIndex: false
        }),
        updateWhileAnimating: true,
        updateWhileInteracting: true
    });

    if (goog.isDef(opt_options)) {
        if (goog.isDef(opt_options.style))
            vectorLayer.setStyle(opt_options.style);

        if (goog.isDef(opt_options.extent))
            vectorLayer.setExtent(opt_options.extent);

        if (goog.isDef(opt_options.minResolution))
            vectorLayer.setMinResolution(opt_options.minResolution);

        if (goog.isDef(opt_options.maxResolution))
            vectorLayer.setMaxResolution(opt_options.maxResolution);

        if (goog.isDef(opt_options.opacity))
            vectorLayer.setOpacity(opt_options.opacity);

        if (goog.isDef(opt_options.properties))
            vectorLayer.setProperties(opt_options.properties);

        if (goog.isDef(opt_options.visible))
            vectorLayer.setVisible(opt_options.visible);

        if (goog.isDef(opt_options.zIndex))
            vectorLayer.setZIndex(opt_options.zIndex);
    }

    // Ajoute la couche à layersToTransform_ pour que les features soient reprojetés en cas de changement de projection
    this.layersToTransform_.push(vectorLayer);

    return vectorLayer;
};

/**
 * Remove events and interactions and then add a interaction
 * @param {ol.interaction.Interaction} interaction
 * @param {string|undefined} currentAction name of the current action
 * @returns {ol.interaction.Interaction}
 */
nsVmap.MapCompare.prototype.setInteraction = function (interaction, currentAction) {
    oVmap.log('nsVmap.MapCompare.prototype.setInteraction');
    // Supprime les anciens évènements et interractions
    this.removeActionsOnMap();
    // Ajoute la nouvelle interaction
    this.addInteraction(interaction);
    // Définit l'action en cours
    if (goog.isDef(currentAction))
        this.setCurrentAction(currentAction);
    return interaction;
};

/**
 * Add a interaction
 * @param {ol.interaction.Interaction} interaction
 * @param {string|undefined} currentAction name of the current action
 * @returns {ol.interaction.Interaction}
 */
nsVmap.MapCompare.prototype.addInteraction = function (interaction, currentAction) {
    oVmap.log('nsVmap.MapCompare.prototype.addInteraction');
    // Ajout de l'interaction
    this.oOpenLayersMap_.addInteraction(interaction);
    // Stocke l'interaction
    this.vmapInteractions_.push(interaction);
    // Définit l'action en cours
    if (goog.isDef(currentAction))
        this.setCurrentAction(currentAction);
    return interaction;
};

/**
 * currentAction setter
 * @param {string} currentAction
 * @export
 */
nsVmap.MapCompare.prototype.setCurrentAction = function (currentAction) {
    var scope = angular.element($("#map1Compare")).scope();
    scope.$evalAsync(function (scope) {
        scope['ctrl']['currentAction'] = currentAction;
    });
};

/**
 * Remove the mapTooltip and the actions added by addEventOnMap, setEventOnMap, addDrawInteraction or setDrawInteraction
 * @returns {undefined}
 * @export
 */
nsVmap.MapCompare.prototype.removeActionsAndTooltips = function () {
    this.removeActionsOnMap();
    //this.getMapTooltip().hide();
    for (var i = 0; i < this.vmapOverlays_.length; i++) {
        this.vmapOverlays_[i].getElement().style.display = "none";
    }
};

/**
 * Remove the events and draw interactions added by addEventOnMap, setEventOnMap, addDrawInteraction or setDrawInteraction
 * @returns {undefined}
 * @export
 */
nsVmap.MapCompare.prototype.removeActionsOnMap = function () {
    oVmap.log('nsVmap.Map.prototype.removeActionsOnMap');
    // Supprime les évènements
    for (var i = 0; i < this.vmapEvents_.length; i++) {
        ol.Observable.unByKey(this.vmapEvents_[i]);
    }

    // Supprime les interractions
    for (var i = 0; i < this.vmapInteractions_.length; i++) {

        // Supprime les features occasionelles
        if (goog.isDef(this.vmapInteractions_[i].getFeatures))
            this.vmapInteractions_[i].getFeatures().clear();
        this.oOpenLayersMap_.removeInteraction(this.vmapInteractions_[i]);
    }

    // Définit l'action en cours
    this.setCurrentAction('');
};

/**
 * Set a event on the map and remove the others.
 * @param {string|Array.<string>} type The event type or array of event types.
 * @param {function(?): ?} listener The listener function.
 * @param {Object=} opt_this The object to use as `this` in `listener`.
 * @param {string|undefined} currentAction name of the current action
 * @return {goog.events.Key} Unique key for the listener.
 * @export
 */
nsVmap.MapCompare.prototype.setEventOnMap = function (type, listener, opt_this, currentAction) {
    oVmap.log('nsVmap.Map.prototype.setEventOnMap');
    if (!goog.isDef(type)) {
        console.error('setEventOnMap: type not defined');
        return 0;
    }

    if (!goog.isDef(listener)) {
        console.error('setEventOnMap: listener not defined');
        return 0;
    }

    opt_this = goog.isDef(opt_this) ? opt_this : '';
    // Supprime les anciens évènements et interractions
    this.removeActionsOnMap();
    // Ajoute le nouvel évènement
    var event = this.addEventOnMap(type, listener, opt_this);
    // Définit l'action en cours
    if (goog.isDef(currentAction))
        this.setCurrentAction(currentAction);
    return event;
};

/**
 * Add a certain type of event on the map.
 * @param {string|Array.<string>} type The event type or array of event types.
 * @param {function(?): ?} listener The listener function.
 * @param {Object=} opt_this The object to use as `this` in `listener`.
 * @param {string|undefined} currentAction name of the current action
 * @return {goog.events.Key} Unique key for the listener.
 * @export
 */
nsVmap.MapCompare.prototype.addEventOnMap = function (type, listener, opt_this, currentAction) {

    if (!goog.isDef(type)) {
        console.error('setEventOnMap: type not defined');
        return 0;
    }

    if (!goog.isDef(listener)) {
        console.error('setEventOnMap: listener not defined');
        return 0;
    }

    opt_this = goog.isDef(opt_this) ? opt_this : '';
    // Active l'évènement
    var event = this.oOpenLayersMap_.on(type, listener, opt_this);
    // Stocke l'évènement
    this.vmapEvents_.push(event);
    // Définit l'action en cours
    if (goog.isDef(currentAction))
        this.setCurrentAction(currentAction);
    return event;
};

/**
 * Get the current scale
 * @param {object} opt_options
 * @param {boolean|undefined} opt_options.pretty true for pretty format (ex: 1:25,000)
 * @returns {Number|String}
 * @export
 */
nsVmap.MapCompare.prototype.getScale = function (opt_options) {
    oVmap.log('nsVmap.MapCompare.prototype.getScale');

    opt_options = goog.isDef(opt_options) ? opt_options : {};
    opt_options['pretty'] = goog.isDef(opt_options['pretty']) ? opt_options['pretty'] : false;

    var wgs84Sphere_ = new ol.Sphere(6378137);
    var projection = this.oOpenLayersMap_.getView().getProjection();

    var map = oVmap.getMapCompare().getOLMap();

    // récupère les coordonnées d'une ligue de 1cm (avec le zoom en cours)
    var line = map.getView().calculateExtent([37.795275591, 0]);
    var c1 = ol.proj.transform([line[0], line[1]], projection, 'EPSG:4326');
    var c2 = ol.proj.transform([line[2], line[3]], projection, 'EPSG:4326');

    // Récuère la longueur sur la carte de la ligne de 1cm
    var length = wgs84Sphere_.haversineDistance(c1, c2);

    // donc 1m sur la carte correspond à (length mètres dans la réalité x 100)
    var scale = length * 100;

    // Rend l'échelle sous format (1:25,000)
    if (opt_options['pretty'] === true) {
        scale = this.getPrettyScale(scale);
    }

    return scale;
};

/**
 * Set the map scale
 * @param {number} scale
 * @returns {number} new scale
 * @export
 */
nsVmap.MapCompare.prototype.setScale = function (scale) {
    oVmap.log('nsVmap.Map.prototype.setScale');

    // calcule et va à l'échelle celon une règle de 3
    var currentScale = this.getScale();
    var currentResolution = this.oOpenLayersMap_.getView().getResolution();

    var scaleResolution = scale * currentResolution / currentScale;
    this.oOpenLayersMap_.getView().setResolution(scaleResolution);

    // Ajuste l'échelle en augementant la résolution
    currentScale = this.getScale();

    oVmap.log(currentScale);
};

/**
 * Get the scale like (1:25,000)
 * @param {number} scale
 * @returns {String} scale (pretty)
 */
nsVmap.MapCompare.prototype.getPrettyScale = function (scale) {

    // Rend l'échelle sous format (1:25,000)
    scale = String(Math.round(scale));
    var j = 1;
    for (var i = scale.length - 1; i > 0; i--) {
        if (j % 3 === 0)
            scale = scale.slice(0, i) + ',' + scale.slice(i + Math.abs(0));
        j++;
    }
    scale = '1:' + scale;

    return scale;
};

/**
 * App-specific directive wrapping the ngeo map directive. The directive's
 * controller has a property "map" including a reference to the OpenLayers
 * map.
 *
 * @return {angular.Directive} The directive specs.
 * @constructor
 */
nsVmap.MapCompare.prototype.mapCompareDirective = function () {
    oVmap.log("nsVmap.MapCompare.prototype.mapCompareDirective");
    return {
        restrict: 'E',
        scope: {
            'map': '=appMap',
            'currentAction': '=appAction'
        },
        controller: 'AppMapCompareController',
        controllerAs: 'ctrl',
        bindToController: true,
        template: '<div id="olMapCompare"></div>'
    };
};

/**
 * The application's main controller.
 * @ngInject
 * @constructor
 */
nsVmap.MapCompare.prototype.mapCompareController = function ($scope, $window, $element, $timeout) {
    oVmap.log("nsVmap.MapCompare.prototype.mapController");
    var this_ = this;

    /**
     * @private
     */
    this.$scope_ = $scope;

    /**
     * @type {object}
     */
    this.element = $element;

    /**
     * Animation duration
     * @type Number
     * @private
     */
    this.duration = 1000;

    /**
     * Timestamp of the last map center sync
     */
    this.lastMapSyncCenter_ = 0;

    /**
     * Timestamp of the last map compare center sync
     */
    this.lastCompareMapSyncCenter_ = 0;

    /**
     * Timestamp of the last map scale sync
     */
    this.lastMapSyncScale_ = 0;

    /**
     * Timestamp of the last map compare scale sync
     */
    this.lastCompareMapSyncScale_ = 0;

    /**
     * @type {ol.Map}
     * @private
     */
    this.map = this['map'];
    /**
     * @type {ol.Overlay}
     * @private
     */
    this.targetOverlay_ = null;

    /**
     * Objet permettant de lancer des fonctions utiles
     * pour parser la définition de la carte
     * @type MapJSON
     */
    var oMapJSON = new MapJSON({
        'properties': oVmap['properties']
    });

    var tileSize = [oVmap['properties']['vmap']['wmsTilesWidth'], oVmap['properties']['vmap']['wmsTilesHeight']];

    /**
     * Contains the tileloaderror and imageloaderror events
     * @private
     */
    this.loadErrorEventsContainer_ = [];

    this.map.setTarget('olMapCompare');

    $scope.$on("mapCompareChanged", function(){
        // MAJ de la taille du composant
        this_.map.setSize([document.getElementById('olMapCompare').offsetWidth, document.getElementById('olMapCompare').offsetHeight]);
        if (this_.map.getLayers().getArray().length > 0)
        this_.map.getLayers().clear();

        var tree_ = oVmap.getMapManager().getCompareLayersTree();
        var olView_ = oMapJSON.getViewFromDef(tree_, {
            'size': this_.map.getSize(),
            'tileSize': tileSize
        });

        var olLayers_ = oMapJSON.getLayersFromDef(tree_, {
            'size': this_.map.getSize(),
            'tileSize': tileSize
        });
        // Définit la vue de la carte
        this_.addView(olView_);
        // Définit les couches de la carte
        for (var i = 0; i < olLayers_.length; i++) {
            this_.addLayer(olLayers_[i]);
        }
        $timeout(function(){
            this_.synchronizeMap();
        }, 10);

        // Rafraichit l'échelle
        setTimeout(function () {
            this_.updateCompareScaleMap();
        });
    })

    /**
     * Update the map size with animation
     * @type goog.async.AnimationDelay
     */
    this.animationDelay = new goog.async.AnimationDelay(
            function () {
                this_.map.updateSize();
                this_.map.renderSync();
                if (goog.now() - start < this_.duration) {
                    this_.animationDelay.start();
                }
            }, $window);

    /**
     * Watch for resizing
     */
    $scope.$watchCollection(function () {
        return [this_.element.children()[0]['clientWidth'], this_.element.children()[0]['clientHeight']];
    }, function () {
        this_.startAnimation();
    });
};

// Définit les directives et controlleurs
oVmap.module.directive('appMapCompare', nsVmap.MapCompare.prototype.mapCompareDirective);
oVmap.module.controller('AppMapCompareController', nsVmap.MapCompare.prototype.mapCompareController);

/**
 * Update the map size with animation
 * @returns {undefined}
 * @export
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.startAnimation = function () {
    oVmap.log('nsVmap.MapCompare.prototype.mapCompareController.prototype.startAnimation');
    start = goog.now();
    this.animationDelay.stop();
    this.animationDelay.start();
};

/**
 * Update the map size with animation
 * @returns {undefined}
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.synchronizeMap = function () {
    oVmap.log('nsVmap.MapCompare.prototype.mapCompareController.prototype.synchronizeMap');

    var this_ = this;

    // Synchro Échelle
    oVmap.getMap().getOLMap().getTargetElement().addEventListener('mouseenter', function(){
        oVmap.getMap().getOLMap().on("moveend", this_.updateCompareScaleMap, this_);
        this_.map.un("moveend", this_.updateScaleMap, this_);
    });
    this_.map.getTargetElement().addEventListener('mouseenter', function(){
      oVmap.getMap().getOLMap().un("moveend", this_.updateCompareScaleMap, this_);
      this_.map.on("moveend", this_.updateScaleMap, this_);
    });

    // Synchro centre
    oVmap.getMap().getOLMap().on("pointerdrag", this_.updateCenterCompareMap, this_);
    this_.map.on("pointerdrag", this_.updateCenterMap, this_);

    // Synchro curseur sur l'autre carte
    oVmap.getMap().getOLMap().on('pointermove', function(e){
      var coord = e.coordinate;
      //si le composant existe sur l'autre composant on le supprime
      if(angular.element(".over-map-compare.ol-target-overlay").length > 0){
        this_.targetOverlay_.setPosition(undefined);
        this_.map.removeOverlay(this_.targetOverlay_);
        this_.targetOverlay_ = undefined;
        angular.element(".over-map-compare.ol-target-overlay").remove();
      }
      // si le composant n'existe pas on le crée
      if (!this_.targetOverlay_){
        var element = document.createElement("div");
    		element.classList.add("ol-target");
    		this_.targetOverlay_ = new ol.Overlay({ 'element': element });
    		this_.targetOverlay_.setPositioning('center-center');
    		this_.map.addOverlay(this_.targetOverlay_);
    		element.parentElement.classList.add("ol-target-overlay");
    		element.parentElement.classList.add("over-map");
    		// hack to render targetOverlay before positioning it
    		this_.targetOverlay_.setPosition([0,0]);
    	}

      var oldProjection = oVmap.getMap().getOLMap().getView().getProjection().getCode();
      var newProjection = this_.map.getView().getProjection().getCode();

      var convertCoord = ol.proj.transform(coord, oldProjection, newProjection)

    	this_.targetOverlay_.setPosition(convertCoord);
    });
    this_.map.on('pointermove', function(e){
      var coord = e.coordinate;
      //si le composant existe sur l'autre composant on le supprime
      if(angular.element(".over-map.ol-target-overlay").length > 0){
        this_.targetOverlay_.setPosition(undefined);
        oVmap.getMap().getOLMap().removeOverlay(this_.targetOverlay_);
        this_.targetOverlay_ = undefined;
        angular.element(".over-map-compare.ol-target-overlay").remove();
      }
      // si le composant n'existe pas on le crée
      if (!this_.targetOverlay_){
        var element = document.createElement("div");
        element.classList.add("ol-target");
        this_.targetOverlay_ = new ol.Overlay({ 'element': element });
        this_.targetOverlay_.setPositioning('center-center');
        oVmap.getMap().getOLMap().addOverlay(this_.targetOverlay_);
        element.parentElement.classList.add("ol-target-overlay");
        element.parentElement.classList.add("over-map-compare");
        // hack to render targetOverlay before positioning it
        this_.targetOverlay_.setPosition([0,0]);
      }

      var oldProjection = this_.map.getView().getProjection().getCode();
      var newProjection = oVmap.getMap().getOLMap().getView().getProjection().getCode();

      var convertCoord = ol.proj.transform(coord, oldProjection, newProjection)

      this_.targetOverlay_.setPosition(convertCoord);
    });
};

/**
 * Update the map extent with animation
 * @returns {undefined}
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.updateCompareScaleMap = function(e) { console.log('updateCompareScaleMap');
    var this_ = this
    this_.updateCenterCompareMap();
    oVmap.getMapCompare().setScale(oVmap.getMap().getScale());

    this.lastCompareMapSyncScale_ = Date.now();
}

/**
 * Update the map extent with animation
 * @returns {undefined}
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.updateScaleMap = function(e) { console.log('updateScaleMap');
    var this_ = this
    this_.updateCenterMap();
    oVmap.getMap().setScale(oVmap.getMapCompare().getScale());

    this.lastMapSyncScale_ = Date.now();
}

/**
* Update the map extent with animation
* @returns {undefined}
*/
nsVmap.MapCompare.prototype.mapCompareController.prototype.updateCenterMap = function(e) {

    var this_ = this
    // Récupère l'étendue de la carte principale
    var oldCenter = this_.map.getView().getCenter();

    // Récupère l'ancienne et la nouvelle projection
    var newProjection = oVmap.getMap().getOLMap().getView().getProjection().getCode();
    var oldProjection = this_.map.getView().getProjection().getCode();

    var newCenter = ol.proj.transform(oldCenter, oldProjection, newProjection);
    oVmap.getMap().getOLMap().getView().setCenter(newCenter);

    this.lastMapSyncCenter_ = Date.now();
}

/**
* Update the map extent with animation
* @returns {undefined}
*/
nsVmap.MapCompare.prototype.mapCompareController.prototype.updateCenterCompareMap = function(e) {

    var this_ = this;
    // Récupère l'étendue de la carte principale
    var oldCenter = oVmap.getMap().getOLMap().getView().getCenter();

    // Récupère l'ancienne et la nouvelle projection
    var oldProjection = oVmap.getMap().getOLMap().getView().getProjection().getCode();
    var newProjection = this_.map.getView().getProjection().getCode();

    var newCenter = ol.proj.transform(oldCenter, oldProjection, newProjection);
    this_['map'].getView().setCenter(newCenter);

    this.lastCompareMapSyncCenter_ = Date.now();
}

/**
 * Add a layer on the map
 * @param {object} olLayer
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.addLayer = function (olLayer) {
    oVmap.log('nsVmap.MapCompare.prototype.mapCompareController.prototype.addLayer');
    this['map'].addLayer(olLayer);
};

/**
 * Add the olView
 * @param {object} olView
 * @returns {undefined}
 */
nsVmap.MapCompare.prototype.mapCompareController.prototype.addView = function (olView) {
    oVmap.log('nsVmap.nsMapManager.LayersTree.prototype.LayertreeController.prototype.addView');

    // Récupère l'étendue de la carte principale
    var oldExtent = oVmap.getMap().getOLMap().getView().calculateExtent(this['map'].getSize());

    // Récupère l'ancienne et la nouvelle projection
    var oldProjection = oVmap.getMap().getOLMap().getView().getProjection().getCode();
    var newProjection = olView.getProjection().getCode();

    // Set le home position pour cette vue
    olView.set('homePosition', {
        center: olView.getCenter(),
        resolution: olView.getResolution()
    });

    oVmap.getMapCompare().setOpenLayersView(olView);
    this['map'].setView(olView);
    this['proj'] = olView.getProjection().getCode();

    // Reprojette en cas de changment de carte
    if (goog.isDef(oldProjection) && goog.isDef(newProjection) && goog.isDef(oldExtent) /*&& goog.isDef(oldPosition)*/) {
        // Reprojette l'étendue
        var newExtent = ol.proj.transformExtent(oldExtent, oldProjection, newProjection);
        this['map'].getView().fit(newExtent, {
            nearest: true
        });
    }
};
